import math
from random import random

import pyglet
import pyglet.window
import pyglet.text
import pyglet.clock

from pyglet.text import document, layout, caret, Label
from pyglet.window import key

from AttributedTextDecoder2 import AttributedTextDecoder2
from ProblemGenerators import *

# Add Emacs keybindings to motion events

for wmod in ('carbon', 'win32', 'xlib'):
    if hasattr(pyglet.window, wmod):
        m = getattr(pyglet.window, wmod)._motion_map
        m[(key.A, key.MOD_CTRL)] = key.MOTION_BEGINNING_OF_LINE
        m[(key.E, key.MOD_CTRL)] = key.MOTION_END_OF_LINE
        m[(key.B, key.MOD_CTRL)] = key.MOTION_LEFT
        m[(key.F, key.MOD_CTRL)] = key.MOTION_RIGHT
        m[(key.H, key.MOD_CTRL)] = key.MOTION_BACKSPACE
        m[(key.D, key.MOD_CTRL)] = key.MOTION_DELETE

        if wmod == 'carbon':
            m[(key.B, key.MOD_OPTION)] = key.MOTION_PREVIOUS_WORD
            m[(key.F, key.MOD_OPTION)] = key.MOTION_NEXT_WORD
    

class Rectangle(object):
    """Draws a rectangle into a batch."""
    def __init__(self, x1, y1, x2, y2, batch):
        self.vertex_list = batch.add(4, pyglet.gl.GL_QUADS, None,
                                     ('v2i', [x1, y1, x2, y1, x2, y2, x1, y2]),
                                     ('c4B', [200, 200, 220, 255] * 4))

class TextEntryWidget(object):
    def __init__(self, text, x, y, width, batch):
        doc = document.UnformattedDocument(text)
        doc.set_style(0, len(doc.text), 
                      dict(color=(0, 0, 0, 255), font_size=24))
        font = doc.get_font()
        height = font.ascent - font.descent

        self.document = doc
        self.layout = layout.IncrementalTextLayout(self.document,
                                    width, height, multiline=False, batch=batch)
        self.caret = caret.Caret(self.layout)

        self.layout.x = x
        self.layout.y = y

        # Rectangular outline
        pad = 3
        self.rectangle = Rectangle(x - pad, y - pad, 
                                   x + width + pad, y + height + pad,
                                   batch)

    def hit_test(self, x, y):
        return (0 < x - self.layout.x < self.layout.width and
                0 < y - self.layout.y < self.layout.height)


class Window(pyglet.window.Window):
    def __init__(self, *args, **kwargs):
        super(Window, self).__init__(600, 600, caption='Answer Test')

        self.batch = pyglet.graphics.Batch()

        black = (0,0,0,255)
        self.base_attributes = dict(color=black, background_color=(255,255,255,255),
                                    font_name='CMU Classical Serif Italic', font_size=36,
                                    baseline=0)
        self.mesg_attributes = dict(color=black, background_color=(255,255,255,255),
                                    font_name='CMU Typewriter Text Regular', font_size=36,
                                    baseline=0)
        self.ansr_attributes = dict(color=black, background_color=(255,255,255,255),
                                    font_name='CMU Concrete Roman', font_size=24,
                                    baseline=0, wrap=True)


        self.answerLabel = Label('Answer', x=10, y=75, anchor_y='bottom',
                                    color=black, font_size=24, batch=self.batch)
        self.answerWidget = TextEntryWidget('', 135, 75, self.width-175, self.batch)

        self.clockTime = 0
        self.clockLabel = Label('Time 00:00.0', x=10, y=10, anchor_y='bottom',
                                    color=black, font_size=24, batch=self.batch)

        self.correct = 0
        self.incorrect = 0
        self.timing = 0
        self.ctiming = 0

        self.content = document.FormattedDocument()
        self.content.insert_text(0, 'Answer each question and hit return.')
        self.content.set_style(0, len(self.content.text), attributes=self.ansr_attributes)
        self.content_layout = layout.TextLayout(self.content, width=(3*self.width)//4, height=(3*self.height)//4, multiline=False, batch=self.batch)
        self.content_layout.x = self.width//2
        self.content_layout.y = self.height//3
        self.content_layout.anchor_x = 'center'
        self.content_layout.anchor_y = 'center'

        self.text_cursor = self.get_system_mouse_cursor('text')

        self.problems = [IntegerMultiplicationProblem([(2,12)]),
                         IntegerAdditionProblem([(1,99)]),
                         PowersOfTwo(-6,12),
                         Squares(0,25),
                         Cubes(-10,10),
                         IntegerSubtractionProblem([(1,99)]),
                         IntegerMultiplicationProblem([(10,50)]),
                         IntegerAdditionProblem([(100,999)]),]

        self.probabilities = [0.3, 0.1, 0.15, 0.1, 0.1, 0.1, 0.05, 0.1]
        self.problem = None
        self.focus = None
        self.scheduled = None
        self.done = False

        pyglet.clock.schedule_interval(self.update_clock, 0.1)
        self.ready_message()

    def update_clock(self, dt):
        if self.focus:
            decis = int(math.floor(10*dt + 0.5))
            time = self.clockTime + decis
            self.clockLabel.text = "Time %02d:%02d.%1d" % ((time//600), (time % 600)//10, (time % 10))
            self.clockTime = time
    
    def on_resize(self, width, height):
        super(Window, self).on_resize(width, height)
        self.answerWidget.width = width - 110

    def on_draw(self):
        pyglet.gl.glClearColor(1, 1, 1, 1)
        self.clear()
        self.batch.draw()


    def on_text(self, text):
        if self.focus:
            self.focus.caret.on_text(text)

    def on_text_motion(self, motion):
        if self.focus:
            self.focus.caret.on_text_motion(motion)
      
    def on_text_motion_select(self, motion):
        if self.focus:
            self.focus.caret.on_text_motion_select(motion)

    def on_key_press(self, symbol, modifiers):
        if symbol == key.RETURN:
            self.prepare_next_question()
        elif symbol == pyglet.window.key.ESCAPE:
            self.finish()
        
    def set_focus(self, focus):
        if self.focus:
            self.focus.caret.visible = False
            self.focus.caret.mark = self.focus.caret.position = 0

        self.focus = focus
        if self.focus:
            self.focus.caret.visible = True
            self.focus.caret.mark = 0
            self.focus.caret.position = len(self.focus.document.text)

    def get_next_problem(self):
        u = random()
        i = 0
        total = self.probabilities[0]

        while i < len(self.problems):
            if u < total:
                break
            i = i + 1
            total += self.probabilities[i]

        self.which_problem = i
        self.problem = self.problems[i].get()

        return (self.problems[i], self.problem)

    def prepare_next_question(self):
        self.content.delete_text(0, len(self.content.text))

        users_answer = self.answerWidget.document.text.strip()
        right_answer = str(self.problems[self.which_problem].solve(self.problem)).strip()

        decoder = AttributedTextDecoder2(self.content, attributes=self.ansr_attributes)

        if users_answer == right_answer:
            response = "Right!"
            self.correct += 1
            self.ctiming += self.clockTime
        else:
            response = "Wrong."
            self.incorrect += 1

        self.timing += self.clockTime

        decoder.decode(response + " The answer is " + right_answer + ".")
        self.answerWidget.document.text = ''
        self.ready_message()

    def ready_message(self):
        self.set_focus(None)

        if not self.done:
            self.scheduled = lambda dt: self.next_question(0)
            pyglet.clock.schedule_once(self.scheduled, 2.0)

        self.clockTime = 0
        self.clockLabel.text = ''

    def next_question(self, state):

        self.content.delete_text(0, len(self.content.text))

        if state < 3:
            msgs = ("{bold True}{color (255,0,0,255)}Ready?",
                    "{bold True}{color (255,255,0,255)}Set...",
                    "{bold True}{color (0,255,0,255)}Go!{color (0,0,0,255)}{bold False} ")

            decoder = AttributedTextDecoder2(self.content, attributes=self.mesg_attributes)
            decoder.decode(msgs[state])

            if not self.done:
                self.scheduled = lambda dt: self.next_question(state+1)
                pyglet.clock.schedule_once(self.scheduled, 0.75)
        else:
            pclass, prob = self.get_next_problem()
            decoder = AttributedTextDecoder2(self.content, attributes=self.base_attributes)
            decoder.decode(pclass.render(prob))

            self.scheduled = None
            self.set_focus(self.answerWidget)
            self.clockTime = 0
            self.clockLabel.text = 'Time 00:00.0'

    def finish(self):
        self.done = True
        if self.scheduled is not None:
            pyglet.clock.unschedule(self.scheduled)

        self.set_focus(None)
        results = "Results %d/%d, Times %.2f/%.2f" % \
                   (self.correct, self.correct + self.incorrect,
                    self.ctiming/max(1,10.0*self.correct),
                    self.timing/max(1,10.0*(self.correct+self.incorrect)))

        self.content.delete_text(0, len(self.content.text))
        decoder = AttributedTextDecoder2(self.content, attributes=self.ansr_attributes)
        decoder.decode(results)
        print results
        
        pyglet.clock.schedule_once( lambda dt: pyglet.app.exit(), 3.0)

        

window = Window(resizable=True) 
pyglet.app.run()
